import string
import re
try:
  from Tkinter import *
  import tkMessageBox
except:
  import kcs_ui
  kcs_ui.message_confirm("Cannot load Tcl or TkInter. They should be available under PYTHONPATH!")
  raise

import kcs_dex

DEXError = "Data Extraction error"

def _getDescription():
  """Helper function for retrieving component descriptions"""
  return string.strip(kcs_dex.get_string())

def _getTypeCode():
  """Helper function for retrieving component type code"""
  return "%07d" % kcs_dex.get_int()

class ComponentBrowser:
  #Mapping of 'wildcards' into Perl5 regexp pattern elements
  reDict = {'*':'.*?', '%':'.', '#':'\\d', '.':'\\.', '$':'\\$',
            '+':'\\+', '\\':'\\\\', '^':'\\^', '{':'\\{',
            '}':'\\}', '[':'\\[', ']':'\\]', '|':'\\|'}
  queryData = [("types", "GEN_PROP.TYPE", 3, kcs_dex.get_string, 0),
               ("type codes", "TYPE", 1, _getTypeCode, 0),
               ("descriptions", "DES(%d)", 3, _getDescription, 1)]
  def __init__(self, master, compType = ""):
    """Creates the Component Browser window. Then performs a search using
    only the component type criterium (if compType is not empty)"""
    self.master = master
    self.nItems = 0
    self.selected = None
    master.title("Component Browser")

    fr1 = Frame(master)
    Vscrollbar = Scrollbar(fr1, orient=VERTICAL)
    Hscrollbar = Scrollbar(fr1, orient=HORIZONTAL)
    self.browser = Listbox(fr1, yscrollcommand=Vscrollbar.set, \
      xscrollcommand=Hscrollbar.set, font="Courier 8", bg="White")
    self.browser.bind("<Double-Button-1>", self.doSelectEvent)
    master.bind("<Return>", self.doSearchEvent)
    Vscrollbar.config(command=self.browser.yview)
    Hscrollbar.config(command=self.browser.xview)
    Hscrollbar.pack(side=BOTTOM, fill=X)
    Vscrollbar.pack(side=RIGHT, fill=Y)
    self.browser.pack(side=LEFT, fill=BOTH, expand=1)
    fr1.pack(side=TOP, fill=BOTH, expand=1)

    fr2 = Frame(master, relief=SUNKEN, borderwidth=1)
    Label(fr2, text="Component selection criteria", fg = "Blue", font = "Helvetica 11 bold").grid(row=0, column=0, columnspan=4)
    Label(fr2, text="Name:").grid(row=1, column=0, sticky=E)
    self.cName = Entry(fr2)
    self.cName.help = "Specify component name (wildcard '*' allowed)"
    self.cName.grid(row=1, column=1)
    Label(fr2, text="Type code:").grid(row=1, column=2, sticky=E)
    self.cTypeCode = Entry(fr2)
    self.cTypeCode.help = "Specify component type code (wildcards '*', '%', '#' allowed)"
    self.cTypeCode.grid(row=1, column=3)
    Label(fr2, text="Description:").grid(row=2, column=0, sticky=E)
    self.cDescr = Entry(fr2)
    self.cDescr.help = "Specify component description (wildcards '*', '%', '#' allowed)"
    self.cDescr.grid(row=2, column=1)
    Label(fr2, text="Type:").grid(row=2, column=2, sticky=E)
    self.cType = Entry(fr2)
    self.cType.help = "Specify component type (wildcards '*', '%', '#' allowed)"
    self.cType.insert(0, string.upper(compType))
    self.cType.grid(row=2, column=3)
    
    fr2a = Frame(fr2)
    butHelp = "Select description language ID (1, 2, or 3)"
    self.langID = IntVar()
    self.langID.set(1)
    b = Radiobutton(fr2a, variable=self.langID, value = 1, command=self.helpLang, text="Language 1")
    b.help = butHelp
    b.pack(side=TOP)
    b = Radiobutton(fr2a, variable=self.langID, value = 2, command=self.helpLang, text="Language 2")
    b.help = butHelp
    b.pack(side=TOP)
    b = Radiobutton(fr2a, variable=self.langID, value = 3, command=self.helpLang, text="Language 3")
    b.help = butHelp
    b.pack(side=TOP)
    fr2a.grid(row=0, column=4, rowspan=3)
    fr2.pack(side=TOP, fill=X, pady=2)

    fr3 = Frame(master)
    b = Button(fr3, text="Search now", command=self.doSearch, width=10)
    b.help = "Click to execute the component search using currently set criteria"
    b.pack(side=LEFT, padx=4)
    self.selButton = Button(fr3, text="Select", command=self.doSelect, width=10, state=DISABLED)
    self.selButton.help = "Click to return with the information about the selected component"
    self.selButton.pack(side=LEFT, padx=4)
    b = Button(fr3, text="Cancel", command=self.doCancel, width=10)
    b.help = "Click to cancel the component selection"
    b.pack(side=LEFT, padx=4)
    fr3.pack(side=TOP, pady=2)

    self.status = Label(master, text="", bd=1, relief=SUNKEN, anchor=W)
    self.status.pack(side=BOTTOM, fill=X)
    master.bind("<FocusIn>", self.help)
    self.browser.bind("<FocusIn>", self.enter_browser)
    self.cName.focus_set()

    self.doSearch() #Performs preliminary search
    master.protocol("WM_DELETE_WINDOW", self.doCancel)
    

  def setStatus(self, msg):
    """Updates the text displayed on the Statusbar"""
    self.status.config(text=msg)

  def enter_browser(self, event):
    """Generates the appropriate text displayed on the Statusbar,
    when the browser takes the focus"""
    if self.nItems > 0:
      self.setStatus("Select a component, and click the Select button")
    else:
      self.setStatus("No components found")
    return "break"

  def helpLang(self):
    self.setStatus("Select description language ID (1, 2, or 3)")
    
  def help(self, event):
    """General event handler for displaying help text on the Statusbar,
    when the widget takes the focus"""
    try:
      msg = event.widget.help
    except:
      msg = "No help available"
    self.setStatus(msg)
    return "break"
    
  def getCompNameMask(self):
    """Generates the canonical component name mask to be used
    in Data Extraction commands"""
    mask = string.strip(self.cName.get())
    if mask == "":
      return "*"
    else:
      lastAlpha, maskLen = -1, len(mask)
      newMask = ""
      for c in mask:
        if c == "*":
          if lastAlpha == 1: #Sequence <alpha>* - insert "'"
            newMask = newMask + "'" + c
          elif lastAlpha != 0: #Sequence ** - ignore additional "*"
            newMask = newMask + c
          lastAlpha = 0
        else:
          if lastAlpha == 1: #Sequence <alpha><alpha> - accept
            newMask = newMask + c
          else: #Sequence *<alpha> or <beg-of-str><alpha> - insert "'"
            newMask = newMask + "'" + c
            lastAlpha = 1
      if lastAlpha:
        newMask = newMask + "'"
    return string.upper(newMask)
    
  def getCompiledRE(self, entry, name, options=0):
    """Prepares the regexp pattern according the contents of the
    given entry, and returns its compiled version, or None on failure."""
    mask = string.strip(entry.get())
    if mask != "":
      reStr = ""
      rd = self.reDict
      for c in mask:
        if rd.has_key(c):
          reStr = reStr + rd[c]
        else:
          reStr = reStr + c
      try:
        RE = re.compile(reStr+"$", options)
      except:
        RE = None
      failed = (RE is None)
      if failed:
        tkMessageBox.showerror("Component Browser", "Invalid component %s specification: '%s'!" % (name, mask))
    else:
      failed, RE = 0, None #Continue, but without this restriction
    return (failed, RE)

  def getCompName(self):
    """Retrieves the component name from the recently executed
    Data Extraction command of type "COMPONENT(...". """
    com = kcs_dex.get_commandstring()
    n = string.find(com, ")", 10)
    return com[10:n]

  def findResult(self, RE, nameSpec, what):
    """Collects the component information from the Data Extraction command of
    the form: COMPONENT('name specification')...
    Returns the list of 4-element lists: [name, type, type code, description]
    
    Parameters:
      RE          - Perl5 regexp pattern for matching the returned values
      nameSpec    - component name specification
      what        - identification of the information to be retrieved"""
    ident, tail, resType, getFunction, withLang = self.queryData[what-1]
    if withLang:
      tail = tail % self.langID.get()
    command = "COMPONENT(%s).%s" % (nameSpec, tail)
    if kcs_dex.extract(command) == 0:
      anyValue = (RE is None)
      result = []
      while kcs_dex.next_result() == resType:
        value = getFunction()
        if anyValue or RE.match(value):
          res = [self.getCompName(), None, None, None]
          res[what] = value
          result.append(res)
      return result
    else:
      raise DEXError, "Error fetching component %s using DEX command: %s" % (ident, command)

  def filterResult(self, result, RE, what):
    """Filters the components currently present in the result list using
    the RE regexp pattern, and information identified by 'what' index.
    
    Returns the filtered list"""
    anyValue = (RE is None)
    if not anyValue: res = []
    ident, tail, resType, getFunction, withLang = self.queryData[what-1]
    if withLang:
      tail = tail % self.langID.get()
    stFormat = "COMPONENT('%s')." + tail
    for item in result:
      st = stFormat % item[0]
      if (kcs_dex.extract(st) == 0) and (kcs_dex.next_result() == resType):
        value = getFunction()
        if anyValue:
          item[what] = value
        elif RE.match(value):
          item[what] = value
          res.append(item)
      else:
        raise DEXError, "Error fetching component %s using DEX command: %s" % \
              (ident, st)
    if anyValue: return result
    else: return res
      
  def doSearch(self):
    """Querries the component database using the criteria given in the
    browser window. Updates the listbox contents to display the result
    of the query."""
    try:
      self.setStatus("Performing component search, please wait ...")
      nameSpec = self.getCompNameMask()
      failed, typeCodeRE = self.getCompiledRE(self.cTypeCode, "type code")
      if failed: return
      failed, descrRE = self.getCompiledRE(self.cDescr, "description", re.I)
      if failed: return
      failed, typeRE = self.getCompiledRE(self.cType, "type", re.I)
      if failed: return

      #Execute the query
      try:
        if typeRE is not None: #component type condition set
          result = self.findResult(typeRE, nameSpec, 1)
          #Filter the result list using type code, and description criteria
          result = self.filterResult(result, typeCodeRE, 2)
          result = self.filterResult(result, descrRE, 3)

        elif typeCodeRE is not None: #type condition not set, type code condition set
          result = self.findResult(typeCodeRE, nameSpec, 2)
          #Filter the result list using description criterium, and add type codes
          result = self.filterResult(result, descrRE, 3)
          result = self.filterResult(result, typeRE, 1)

        elif descrRE is not None: #type and type code conditions not set, description condition set
          result = self.findResult(descrRE, nameSpec, 3)
          #Add type codes and types
          result = self.filterResult(result, typeRE, 1)
          result = self.filterResult(result, typeCodeRE, 2)
        else:                                              #possibly only name condition set
          if nameSpec == "*":
            if tkMessageBox.askquestion("Component Browser", \
              "No conditions set! Do you really want to retrieve ALL components?") != tkMessageBox.YES: return
          #No conditions set, except component name. Find all matching type codes,
          #and add type and description information
          result = self.findResult(typeCodeRE, nameSpec, 2)
          result = self.filterResult(result, descrRE, 3)
          result = self.filterResult(result, typeRE, 1)
      except DEXError, msg:
        tkMessageBox.showerror("Component Browser", msg)
        return
        
      #Update the browser contents      
      list = self.browser
      list.delete(0, END)
      self.nItems = len(result)
      if self.nItems == 0:
        list.insert(END, "No component matches the query criteria!")
        self.selButton.config(state=DISABLED)
        self.master.title("Component Browser")
      else:
        self.widths = reduce(lambda x, y: (max(x[0], len(y[0])), max(x[1], len(y[1]))), result, (1, 1))
        format = "%%-%ds  %%-%ds  %%s  %%s" % self.widths
        for item in result:
          list.insert(END, format % tuple(item))
        self.selButton.config(state=NORMAL)
        self.master.title("Component Browser (%d items found)" % self.nItems)
    finally:
      self.enter_browser(None)
      
  def doSelect(self):
    res = self.browser.curselection()
    if (self.nItems > 0) and (len(res) > 0):
      item = self.browser.get(res[0])
      eName=self.widths[0]
      eType=eName+self.widths[1]+2
      eTypeCode=eType+9
      self.selected = [string.strip(item[:eName]), string.strip(item[eName+2:eType]), \
                       string.strip(item[eType+2:eTypeCode]), string.strip(item[eTypeCode+2:])]

      self.master.quit()

  def doSearchEvent(self, event):
    self.doSearch()

  def doSelectEvent(self, event):
    self.doSelect()    

  def doCancel(self):
    self.selected = None
    self.master.quit()

def run(compType=""):
  """run() -> [name, type, type code, description], or None
  
  Displays the Component Browser window, and lets the user make querries to the
  General Component Database. If the user chooses the component, the relevant
  information is returned in the 4-element list. If the user cancels the selection,
  None is returned"""
  root = Tk()
  app = ComponentBrowser(root, compType)
  root.mainloop()
  res = app.selected
  root.destroy()
  return res

if __name__ == "__main__":
  root = Tk()
  app = ComponentBrowser(root, "ACC1")
  root.mainloop()
  res = app.selected
  if res is None:
    msg = "No component has been selected!"
  else:
    print res
    msg = "The following component has been selected:\nName: %s\nType: %s\nType code: %s\nDescription: %s" % \
          tuple(res)
  tkMessageBox.showinfo("Component Browser", msg)
  root.destroy()
  
    